module rs232_debugger (

    input  wire clk,

    input  wire rst,
	output reg  cmd_rst,   // When sent by the debugger utility, this output pulses for 8 clock cycles

    input  wire rxd,
    output wire txd,
	output reg         LED_txd,
	output reg         LED_rxd,

	output reg         host_rd_req,
	output reg         host_wr_ena,
	output wire [23:0] host_addr,
	output reg  [7:0]  host_wdata,
	input  wire [7:0]  host_rdata,
	
	input  wire [7:0]  in0,
	input  wire [7:0]  in1,
	input  wire [7:0]  in2,
	input  wire [7:0]  in3,
	output reg  [7:0]  out0,
	output reg  [7:0]  out1,
	output reg  [7:0]  out2,
	output reg  [7:0]  out3
	
	);


parameter CLK_IN_HZ    = 50000000;
parameter BAUD_RATE    = 921600;
parameter DEBUG_RST_ADR = 0;	// sets read/write address after a reset pulse

localparam DEBUG_TIMEOUT = 24;   // Time until a incomming command is flushed due to inactive com. Time in seconds = ( 2^DEBUG_TIMEOUT / CLK_IN Hz )
localparam LED_HOLD_TIME = 24;   // Keep the LED_rxd/txd signal high for this amount of time during a RXD/TXD transaction. Time in seconds = ( 2^LED_HOLD_TIME / CLK_IN Hz )

localparam ECHO_DATA = 1 ;		// For write verify, this features echoing the writen data back out of the RS232 port.


reg [23:0] host_addr_reg    = DEBUG_RST_ADR;       // Used to assign a power-up default
assign     host_addr[23:0]  = host_addr_reg[23:0];
wire [7:0] uart_rbyte;
wire       rxd_rdy;
reg        ena_rxd;
reg        [4:0]    ena_rxd_dly;
reg        timeout, h_lock;
reg        [15*8+7:0]  command;
reg		   [3:0]       command_00_cnt;
reg		   [3:0]       command_FF_cnt;
reg        [1:0]      func;
reg        [DEBUG_TIMEOUT:0]     timeout_cnt;
reg        [LED_HOLD_TIME:0]     led_txd_timeout, led_rxd_timeout;
reg        [3:0]      tx_cyc;
reg        [16:0]     byte_count;
reg                   uart_tx;
wire                  uart_tx_full;
reg        [3:0]      rst_clk;
reg        [7:0]      uart_tbyte;
reg                   rxd_reg;
reg		   [7:0]	  in_reg0,in_reg1,in_reg2,in_reg3;


sync_rs232_uart  rs232_io (	.clk(clk),
								.rxd(rxd),
								.txd(txd),

								.rx_data(uart_rbyte),
								.rx_rdy(rxd_rdy),

								.ena_tx(uart_tx),
								.tx_data(uart_tbyte),
								.tx_busy(uart_tx_full) );
	defparam
		rs232_io.CLK_IN_HZ    = CLK_IN_HZ,
		rs232_io.BAUD_RATE    = BAUD_RATE;


always @ (posedge clk) begin


// ******************************************************************
// ****** Generate a status activity RXD and TXD led driver output.
// ****** This routine keeps the LED outputs on long enough to
// ****** visibly see as the data bursts are too short to be seen.
// ******************************************************************
	if (~txd) led_txd_timeout<=0;
	else if (~led_txd_timeout[LED_HOLD_TIME]) led_txd_timeout<=led_txd_timeout+1;
	LED_txd <= ~led_txd_timeout[LED_HOLD_TIME];

	rxd_reg <= rxd;  // Personal preference, I prefer that a input pin doesn't directly feed combinational logic, so I clock register that input
	if (~rxd_reg) led_rxd_timeout<=0;
	else if (~led_rxd_timeout[LED_HOLD_TIME]) led_rxd_timeout<=led_rxd_timeout+1;
	LED_rxd <= ~led_rxd_timeout[LED_HOLD_TIME];
// ******************************************************************
// ******************************************************************


// *** Generate an 8 clock wide reset pulse
cmd_rst <= ~rst_clk[3] ;
if (~rst_clk[3]) begin 
				rst_clk <= rst_clk + 1;
end else begin

if (rst) begin
		host_addr_reg   <= DEBUG_RST_ADR ;
		host_wr_ena     <= 0 ;
		ena_rxd_dly     <= 0 ;
		command[15*8+7:0]         <= 0 ;
		h_lock          <= 0 ;
		timeout_cnt     <= 0 ;
		tx_cyc  		<= 15 ;
		func            <= 0 ;
		uart_tx			<= 0 ;
		end else begin


// *** setup a delayed pipe for the read strobe ena_rxd allowing the read fifo contents to be made ready
			ena_rxd_dly[0]    <=  rxd_rdy ;
			ena_rxd_dly[4:1]  <=  ena_rxd_dly[3:0] ;


if (func[1:0]==0) begin
			host_rd_req <= 0;
			host_wr_ena <= 0;
			
					if (ena_rxd_dly[4]) uart_tbyte    <= uart_rbyte ;

		if (ena_rxd_dly[3]) begin  // Note we are using the ena_rxd_dly[2] since the write ram function will begin filling the write enable too early if the function comes in before the write pulse
							command[7:0]         <= uart_rbyte;
							command[15*8+7:8]    <= command[14*8+7:0];

							if ( command[15*8+7:15*8]==8'h00 ) 		command_00_cnt <= command_00_cnt + (command_00_cnt!=7);
							else 									command_00_cnt <= 0;
							if ( command[15*8+7:15*8]==8'hFF )		command_FF_cnt <= command_FF_cnt + (command_FF_cnt!=7);
							else if ( command[15*8+7:15*8]!=8'h00 )	command_FF_cnt <= 0;

							end

if ( command_00_cnt==6 && command_FF_cnt==2 &&  command[15*8+7:8*8]==command[7*8+7:0*8]  )begin

				if  ( command[7*8+7:4*8] == 32'h52656164 ) begin  // read host ram and transmit to RS232, from 1 through 256 bytes
													byte_count[16:8]    <= 0;
													byte_count[7:0]     <= command[7:0];
													host_addr_reg[23:0] <= command[8+23:8];
													func                <= 1;
													tx_cyc			    <= 0;
													command[15*8+7:0]   <= 128'hFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF;

		end else if ( command[7*8+7:4*8] == 32'h57726974 ) begin    // Read RS232 data and write into host ram, from 1 through 256 bytes
													byte_count[16:8]    <= 0;
													byte_count[7:0]     <= command[7:0];
													host_addr_reg[23:0] <= command[8+23:8];
													func                <= 2;
													command[15*8+7:0]   <= 128'hFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF;
													timeout_cnt         <= 0;  // clear recieve abort timeout counter

		end else if  ( command[7*8+7:4*8] == 32'h52656150 ) begin  // Page read host ram and transmit to RS232, from 256 through 65536 bytes
													byte_count[16]      <= 0;
													byte_count[15:8]    <= command[7:0];
													byte_count[7:0]     <= 8'hFF ;
													host_addr_reg[23:0] <= command[8+23:8];
													func                <= 1;
													tx_cyc			    <= 0;
													command[15*8+7:0]   <= 128'hFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF;

		end else if ( command[7*8+7:4*8] == 32'h57726950 ) begin    // Page Read RS232 data and write into host ram, from 256 through 65536 bytes
													byte_count[16]      <= 0;
													byte_count[15:8]    <= command[7:0];
													byte_count[7:0]     <= 8'hFF ;
													host_addr_reg[23:0] <= command[8+23:8];
													func                <= 2;
													command[15*8+7:0]   <= 128'hFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF;
													timeout_cnt         <= 0;  // clear recieve abort timeout counter

		end else if ( command[7*8+7:0*8] == 64'h52657365744E6F77  ) begin    // Software system wide external reset trigger
													rst_clk             <= 0;
													func                <= 0;
													command[15*8+7:0]   <= 128'hFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF;

		end else if ( command[7*8+7:4*8] == 32'h53657450		  ) begin    // Set the general purpose output ports and read general purpose input ports
													out0				<= command[3*8+7:3*8];
													out1				<= command[2*8+7:2*8];
													out2				<= command[1*8+7:1*8];
													out3				<= command[0*8+7:0*8];
													// *** Parallel register input debug ports.
													in_reg0 <= in0;
													in_reg1 <= in1;
													in_reg2 <= in2;
													in_reg3 <= in3;

													byte_count[16]      <= 0;
													byte_count[15:0]    <= 3;
													func                <= 3;
													tx_cyc			    <= 0;
													command[15*8+7:0]   <= 128'hFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF;


		end 
	end // Command verification.

end // all of func[1:0]==0


	if (func[1:0]==1) begin   // read host ram and transmit to RS232
			if (~byte_count[16]) begin  //  keep on transmitting until byte counter elapses

			if (~(uart_tx_full && tx_cyc==0)) begin
				tx_cyc <= tx_cyc + 1;
					if (tx_cyc==0) host_rd_req <= 1;		// pulse the host_rd_req with the current valid address
					else           host_rd_req <= 0;
					if (tx_cyc==13) uart_tbyte <= host_rdata; // expect the returned data ready within 12 clock cycles, and latch that data into the RS232 transmitter data input register
					if (tx_cyc==14) uart_tx <= 1;			  // Pulse the RS232 transmitter.
					else            uart_tx <= 0;
					if (tx_cyc==15) host_addr_reg   <= host_addr_reg + 1;
					if (tx_cyc==15) byte_count      <= byte_count - 1;
				end

			end else begin
			func        <= 0;
			uart_tx     <= 0;
			host_rd_req <= 0;
			end
	end

	if (func[1:0]==2) begin  										   // Read RS232 data and write into host ram.
			if (~byte_count[16] && ~timeout_cnt[DEBUG_TIMEOUT] ) begin  //  keep on reading from RS232 until byte counter elapses, or the timeout counter has reached it's end

					if (ena_rxd_dly[0]) timeout_cnt <= 0;              // If a character is received, reset the timeout counter
					else                timeout_cnt <= timeout_cnt +1; // when not receiving characters, increment the timeout counter

					if (ena_rxd_dly[0]) host_wdata    <= uart_rbyte ;
					if (ena_rxd_dly[0]) uart_tbyte    <= uart_rbyte ;

					if (ena_rxd_dly[1])   host_wr_ena <= 1;
					else                  host_wr_ena <= 0;

					if (ena_rxd_dly[1])   uart_tx <= 1;					// echo back received characters
					else                  uart_tx <= 0;

					if (ena_rxd_dly[3]) host_addr_reg   <= host_addr_reg + 1;
					if (ena_rxd_dly[3]) byte_count      <= byte_count - 1;

			end else begin
			func <= 0;
			host_wr_ena   <= 0;
			byte_count[16] <= 1;
			end
	end	



	if (func[1:0]==3) begin   // read general purpose input ports and transmit to RS232
			if (~byte_count[16]) begin  //  keep on transmitting until byte counter elapses

			if (~(uart_tx_full && tx_cyc==0)) begin
				tx_cyc <= tx_cyc + 1;

					if (tx_cyc==13) begin
						if (byte_count==3) uart_tbyte	<= in_reg0;
						else if (byte_count==2) uart_tbyte	<= in_reg1;
						else if (byte_count==1) uart_tbyte	<= in_reg2;
						else if (byte_count==0) uart_tbyte	<= in_reg3;
					end

					if (tx_cyc==14) uart_tx <= 1;			  // Pulse the RS232 transmitter.
					else            uart_tx <= 0;
					if (tx_cyc==15) byte_count      <= byte_count - 1;
				end

			end else begin
			func        <= 0;
			uart_tx     <= 0;
			host_rd_req <= 0;
			end
	end


 end // !rst

end // !soft_rst
end // always

endmodule

